"""Test deltacycle.queue"""

# pyright: reportAttributeAccessIssue=false
# pyright: reportUnknownMemberType=false
# pyright: reportUnknownVariableType=false

import logging

from pytest import LogCaptureFixture

from deltacycle import Queue, create_task, run, sleep

logger = logging.getLogger("deltacycle")


EXP1 = {
    (0, "P", "0"),
    (0, "C", "0"),
    (10, "P", "1"),
    (10, "C", "1"),
    (20, "P", "2"),
    (20, "C", "2"),
    (30, "P", "3"),
    (30, "C", "3"),
    (40, "P", "4"),
    (40, "C", "4"),
    (50, "P", "5"),
    (50, "C", "5"),
    (60, "P", "6"),
    (60, "C", "6"),
    (70, "P", "7"),
    (70, "C", "7"),
    (80, "P", "8"),
    (80, "C", "8"),
    (90, "P", "9"),
    (90, "C", "9"),
    (100, "P", "CLOSED"),
}


def test_prod_cons1(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    q: Queue[int] = Queue()

    async def prod():
        for i in range(10):
            logger.info("%d", i)
            await q.put(i)
            await sleep(10)
        logger.info("CLOSED")

    async def cons():
        while True:
            i = await q.get()
            logger.info("%d", i)

    async def main():
        create_task(prod(), name="P")
        create_task(cons(), name="C")

    run(main())

    msgs = {(r.time, r.taskName, r.getMessage()) for r in caplog.records}
    assert msgs == EXP1


EXP2 = {
    (0, "P", "0"),
    (0, "P", "1"),
    (0, "P", "2"),
    (10, "C", "0"),
    (10, "P", "3"),
    (20, "C", "1"),
    (20, "P", "4"),
    (30, "C", "2"),
    (30, "P", "5"),
    (40, "C", "3"),
    (40, "P", "6"),
    (50, "C", "4"),
    (50, "P", "7"),
    (60, "C", "5"),
    (60, "P", "8"),
    (70, "C", "6"),
    (70, "P", "9"),
    (80, "C", "7"),
    (80, "P", "CLOSED"),
    (90, "C", "8"),
    (100, "C", "9"),
}


def test_prod_cons2(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    q: Queue[int] = Queue(2)

    assert q.capacity == 2

    async def prod():
        for i in range(10):
            logger.info("%d", i)
            await q.put(i)
        logger.info("CLOSED")

    async def cons():
        while True:
            await sleep(10)
            i = await q.get()
            logger.info("%d", i)

    async def main():
        create_task(prod(), name="P")
        create_task(cons(), name="C")

    run(main())

    msgs = {(r.time, r.taskName, r.getMessage()) for r in caplog.records}
    assert msgs == EXP2


def test_prod_cons3():
    q: Queue[int] = Queue(2)
    assert len(q) == 0

    async def prod():
        assert q.try_put(1)
        assert len(q) == 1

        assert q.try_put(2)
        assert len(q) == 2

        assert not q.try_put(3)

    async def cons():
        await sleep(10)

        success, value = q.try_get()
        assert success
        assert value == 1

        success, value = q.try_get()
        assert success
        assert value == 2

        success, value = q.try_get()
        assert not success

    async def main():
        create_task(prod())
        create_task(cons())

    run(main())



"""Task: coroutine wrapper"""

from __future__ import annotations

import heapq
from abc import ABC, abstractmethod
from collections import Counter, OrderedDict, deque
from collections.abc import Coroutine, Generator
from enum import IntEnum
from types import TracebackType
from typing import Any, Self

from ._kernel_if import KernelIf

type TaskCoro = Coroutine[None, Sendable | None, Any]
type TaskArgs = tuple[Task.Command] | tuple[Task.Command, Sendable | Signal]


class Signal(Exception):
    """Throw a signal to a task."""


class Interrupt(Signal):
    """Interrupt task."""


class _Kill(Signal):
    """Kill task."""


class TaskQueue(ABC):
    @abstractmethod
    def __bool__(self) -> bool:
        """Return True if the queue has tasks ready to run."""

    @abstractmethod
    def push(self, item: Any) -> None:
        """Push item to queue tail."""

    @abstractmethod
    def pop(self) -> Any:
        """Pop item from queue head."""

    @abstractmethod
    def drop(self, task: Task) -> None:
        """Drop task from queue."""


class EventQ(TaskQueue):
    """Tasks wait for event trigger."""

    def __init__(self):
        self._tasks = OrderedDict[Task, None]()
        self._items = deque[Task]()

    def __bool__(self) -> bool:
        return bool(self._items)

    def push(self, item: Task):
        item.link(self)
        self._tasks[item] = None

    def pop(self) -> Task:
        task = self._items.popleft()
        self.drop(task)
        return task

    def drop(self, task: Task):
        del self._tasks[task]
        task.unlink(self)

    def load(self):
        assert not self._items
        self._items.extend(self._tasks)


class SemaphoreQ(TaskQueue):
    """Tasks wait for a slot to become available."""

    def __init__(self):
        # priority, index, task
        self._items = list[tuple[int, int, Task]]()

        # Monotonically increasing integer
        # Breaks (time, priority, ...) ties in the heapq
        self._index: int = 0

    def __bool__(self) -> bool:
        return bool(self._items)

    def push(self, item: tuple[int, Task]):
        priority, task = item
        task.link(self)
        heapq.heappush(self._items, (priority, self._index, task))
        self._index += 1

    def pop(self) -> Task:
        _, _, task = heapq.heappop(self._items)
        task.unlink(self)
        return task

    def _find(self, task: Task) -> int:
        for i, (_, _, t) in enumerate(self._items):
            if t is task:
                return i
        assert False  # pragma: no cover

    def drop(self, task: Task):
        index = self._find(task)
        self._items.pop(index)
        task.unlink(self)


class CreditQ(TaskQueue):
    """Tasks wait for credit to become available."""

    def __init__(self):
        # priority, index, task, n
        self._items = list[tuple[int, int, Task, int]]()

        # Monotonically increasing integer
        # Breaks (time, priority, ...) ties in the heapq
        self._index: int = 0

    def __bool__(self) -> bool:
        return bool(self._items)

    def push(self, item: tuple[int, Task, int]):
        priority, task, n = item
        task.link(self)
        heapq.heappush(self._items, (priority, self._index, task, n))
        self._index += 1

    def pop(self) -> tuple[Task, int]:
        _, _, task, n = heapq.heappop(self._items)
        task.unlink(self)
        return task, n

    def _find(self, task: Task) -> int:
        for i, (_, _, t, _) in enumerate(self._items):
            if t is task:
                return i
        assert False  # pragma: no cover

    def drop(self, task: Task):
        index = self._find(task)
        self._items.pop(index)
        task.unlink(self)

    def peek(self) -> int:
        assert self._items
        return self._items[0][-1]


class Blocking(ABC):
    @abstractmethod
    def try_block(self, task: Task) -> bool:
        """Attempt to block task; return True if successful."""

    @property
    @abstractmethod
    def s(self) -> Sendable:
        """Object that will be sent to unblock task."""


class Sendable(ABC):
    @abstractmethod
    def wait_drop(self, task: Task) -> None:
        """Drop task from object's waiting queue."""


class _SuspendResume:
    """Suspend/Resume current task.

    Use case:
    1. Current task A suspends itself: RUNNING => WAITING
    2. Kernel chooses PENDING tasks ..., T
    3. ... Task T wakes up task A w/ value X: WAITING => PENDING
    4. Kernel chooses PENDING tasks ..., A: PENDING => RUNNING
    5. Task A resumes with value X

    The value X can be used to pass information to the task.
    """

    def __await__(self) -> Generator[None, Sendable | None, Sendable | None]:
        # Suspend
        value = yield
        # Resume
        return value


class _Condition(KernelIf):
    def __init__(self, *bs: Blocking):
        self._bs = bs


class AllOf(_Condition):
    def __await__(self) -> Generator[None, Sendable, tuple[Sendable, ...]]:
        task = self._kernel.task()

        blocked = set[Sendable]()
        unblocked = deque[Sendable]()

        for b in self._bs:
            if b.try_block(task):
                blocked.add(b.s)
            else:
                unblocked.append(b.s)

        while blocked:
            s = yield from task.switch_gen()
            blocked.remove(s)
            unblocked.append(s)

        return tuple(unblocked)


class AnyOf(_Condition):
    def __await__(self) -> Generator[None, Sendable, Sendable | None]:
        if not self._bs:
            return None

        task = self._kernel.task()

        blocked = set[Sendable]()

        for b in self._bs:
            if b.try_block(task):
                blocked.add(b.s)
            else:
                while blocked:
                    s = blocked.pop()
                    s.wait_drop(task)
                return b.s

        self._kernel.fork(task, *blocked)
        s = yield from task.switch_gen()
        return s


class Task(KernelIf, Blocking, Sendable):
    """Manage the life cycle of a coroutine.

    Do NOT instantiate a Task directly.
    Use ``create_task`` function, or (better) ``TaskGroup.create_task`` method.
    """

    class Command(IntEnum):
        START = 0b00
        RESUME = 0b01
        SIGNAL = 0b10

    class State(IntEnum):
        """
        Transitions::

                    PENDING
                       |
            INIT -> RUNNING -> RETURNED
                            -> EXCEPTED
        """

        # Initialized
        INIT = 0b001

        # Currently running
        RUNNING = 0b010

        # Suspended
        PENDING = 0b011

        # Done: returned a result
        RETURNED = 0b100
        # Done: raised an exception
        EXCEPTED = 0b101

    _done = State.RETURNED & State.EXCEPTED

    _state_transitions = {
        State.INIT: {
            State.RUNNING,
        },
        State.RUNNING: {
            State.PENDING,
            State.RETURNED,
            State.EXCEPTED,
        },
        State.PENDING: {
            State.RUNNING,
        },
    }

    def __init__(
        self,
        coro: TaskCoro,
        name: str,
    ):
        self._state = self.State.INIT

        # Attributes
        self._coro = coro
        self._name = name

        # Set if created within a group
        self._group: TaskGroup | None = None

        # Keep track of all queues containing this task
        self._refcnts = Counter[TaskQueue]()

        # Other tasks waiting for this task to complete
        self._waiting = EventQ()

        # Flag to avoid multiple signals
        self._signal = False

        # Outputs
        self._result: Any = None
        self._exception: Exception | None = None

    def _blocking(self) -> bool:
        return not self.done()

    def wait_push(self, task: Task):
        self._waiting.push(task)

    def wait_drop(self, task: Task):
        self._waiting.drop(task)

    def __await__(self) -> Generator[None, Sendable, Any]:
        if self._blocking():
            task = self._kernel.task()
            self.wait_push(task)
            t = yield from task.switch_gen()
            assert t is self

        return self.result()

    @property
    def coro(self) -> TaskCoro:
        """Wrapped coroutine."""
        return self._coro

    @property
    def name(self) -> str:
        """Task name.

        Primarily for debug; no functional effect.
        There are no rules or restrictions for valid names.
        Give tasks unique and recognizable names to help identify them.

        If not provided to the create_task function,
        a default name of ``Task-{index}`` will be assigned,
        where ``index`` is a monotonically increasing integer value,
        starting from 0.
        """
        return self._name

    def _get_group(self) -> TaskGroup | None:
        """Return TaskGroup, or None.

        If the task was started by a TaskGroup's create_task method,
        it will assign this property to point to the TaskGroup instance.
        """
        return self._group

    def _set_group(self, group: TaskGroup):
        self._group = group

    group = property(fget=_get_group, fset=_set_group)

    def _set_state(self, state: State):
        assert state in self._state_transitions[self._state]
        self._state = state

    def state(self) -> State:
        return self._state

    def link(self, tq: TaskQueue):
        self._refcnts[tq] += 1

    def unlink(self, tq: TaskQueue):
        assert self._refcnts[tq] > 0
        self._refcnts[tq] -= 1

    def _renege(self):
        tqs = set(self._refcnts.keys())
        while tqs:
            tq = tqs.pop()
            while self._refcnts[tq]:
                tq.drop(self)
            del self._refcnts[tq]

    async def switch_coro(self) -> Sendable | None:
        self._set_state(Task.State.PENDING)

        # Suspend
        value = await _SuspendResume()
        # Resume
        return value

    def switch_gen(self) -> Generator[None, Sendable, Sendable]:
        self._set_state(self.State.PENDING)

        # Suspend
        value = yield
        # Resume
        return value

    def do_run(self, args: TaskArgs):
        self._set_state(self.State.RUNNING)

        match args:
            case (self.Command.START,):
                y = self._coro.send(None)
            case (self.Command.RESUME,):
                y = self._coro.send(None)
            case (self.Command.RESUME, Sendable() as s):
                y = self._coro.send(s)
            case (self.Command.SIGNAL, Signal() as s):
                self._signal = False
                y = self._coro.throw(s)
            case _:  # pragma: no cover
                assert False

        # TaskCoro YieldType=None
        assert y is None

    def _set(self):
        self._waiting.load()

        while self._waiting:
            task = self._waiting.pop()
            self._kernel.join_any(task, self)
            self._kernel.call_soon(task, args=(self.Command.RESUME, self))

    def do_result(self, exc: StopIteration):
        self._result = exc.value
        self._set_state(self.State.RETURNED)
        self._set()
        assert self._refcnts.total() == 0

    def do_except(self, exc: Exception):
        self._exception = exc
        self._set_state(self.State.EXCEPTED)
        self._set()
        assert self._refcnts.total() == 0

    def done(self) -> bool:
        """Return True if the task is done.

        A task that is "done" either:

        * Completed normally, or
        * Raised an exception.
        """
        return bool(self._state & self._done)

    def result(self) -> Any:
        """Return the task's result, or raise an exception.

        Returns:
            If the task ran to completion, return its result.

        Raises:
            Exception: If the task raise any other type of exception.
            RuntimeError: If the task is not done.
        """
        if self._state is self.State.RETURNED:
            assert self._exception is None
            return self._result
        if self._state is self.State.EXCEPTED:
            assert self._result is None and self._exception is not None
            raise self._exception
        raise RuntimeError("Task is not done")

    def exception(self) -> Exception | None:
        """Return the task's exception.

        Returns:
            If the task raised an exception, return it.
            Otherwise, return None.

        Raises:
            RuntimeError: If the task is not done.
        """
        if self._state is self.State.RETURNED:
            assert self._exception is None
            return self._exception
        if self._state is self.State.EXCEPTED:
            assert self._result is None and self._exception is not None
            return self._exception
        raise RuntimeError("Task is not done")

    def interrupt(self, *args: Any) -> bool:
        """Interrupt task.

        If a task is already done: return False.

        If a task is pending:

        1. Renege from all queues
        2. Reschedule to raise Interrupt in the current time slot
        3. Return True

        If a task is running, immediately raise Interrupt.

        Args:
            args: Arguments passed to Interrupt instance

        Returns:
            bool success indicator

        Raises:
            Interrupt: If the task interrupts itself
        """
        # Already done; do nothing
        if self._signal or self.done():
            return False

        irq = Interrupt(*args)

        # Task is interrupting itself. Weird, but legal.
        if self is self._kernel.task():
            raise irq

        # Pending tasks must first renege from queues
        self._renege()

        # Reschedule
        self._signal = True
        self._kernel.call_soon(self, args=(self.Command.SIGNAL, irq))

        # Success
        return True

    def kill(self) -> bool:
        # Already done; do nothing
        if self._signal or self.done():
            return False

        # Task cannot kill itself
        assert self is not self._kernel.task()

        # Pending tasks must first renege from queues
        self._renege()

        # Reschedule
        self._signal = True
        self._kernel.call_soon(self, args=(self.Command.SIGNAL, _Kill()))

        # Success
        return True

    # Blocking
    def try_block(self, task: Task) -> bool:
        if self._blocking():
            self.wait_push(task)
            return True
        return False

    @property
    def s(self) -> Task:
        return self


class TaskGroup(KernelIf):
    """Group of tasks."""

    def __init__(self):
        self._parent = self._kernel.task()

        # Tasks started in the with block
        self._setup_done = False
        self._setup_tasks = set[Task]()

        # Tasks in running/pending/killing state
        self._todo = set[Task]()

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: type[Exception] | None,
        exc: Exception | None,
        traceback: TracebackType | None,
    ):
        self._setup_done = True

        # Start newly created tasks; ignore exceptions handled by parent
        while self._setup_tasks:
            child = self._setup_tasks.pop()
            if not child.done():
                child.wait_push(self._parent)
                self._todo.add(child)

        # Parent raised an exception:
        # Kill children; suppress exceptions
        if exc:
            for child in self._todo:
                child.kill()
            while self._todo:
                child = await self._parent.switch_coro()
                assert isinstance(child, Task)
                self._todo.remove(child)

            # Re-raise parent exception
            return False

        # Parent did NOT raise an exception:
        # Await children; collect exceptions
        child_excs = list[Exception]()
        killed = set[Task]()
        while self._todo:
            child = await self._parent.switch_coro()
            assert isinstance(child, Task)
            self._todo.remove(child)
            if child in killed:
                continue
            exc = child.exception()
            if exc is not None:
                child_excs.append(exc)
                killed.update(c for c in self._todo if c.kill())

        # Re-raise child exceptions
        if child_excs:
            raise ExceptionGroup("Child task(s) raised exception(s)", child_excs)

    def create_task(
        self,
        coro: TaskCoro,
        name: str | None = None,
        **kwargs: Any,
    ) -> Task:
        child = self._kernel.create_task(coro, name, **kwargs)
        child.group = self
        if self._setup_done:
            if not child.done():
                child.wait_push(self._parent)
                self._todo.add(child)
        else:
            self._setup_tasks.add(child)
        return child



"""Execution Kernel"""

import heapq
from abc import ABC, abstractmethod
from collections.abc import Generator
from enum import IntEnum
from typing import Any, override
from weakref import WeakKeyDictionary

from ._task import Sendable, Task, TaskArgs, TaskCoro, TaskQueue
from ._variable import Variable


class _Finish(Exception):
    """Force the simulation to stop."""


class Kernel(ABC):
    """Simulation Kernel.

    Responsible for:

    * Scheduling and executing tasks
    * Updating model state

    Time is an integer.
    Its initial value is -1, indicating "before time".
    All tasks must be scheduled at a non-negative time.

    All kernels have a ``main`` task,
    which is scheduled to run at time zero.

    This is a low level API.
    User code is not expected to interact with it directly.
    To run a simulation, use the ``run`` and ``step`` functions.
    """

    _index = 0

    class State(IntEnum):
        """
        Transitions::

            INIT -> RUNNING -> COMPLETED
                            -> FINISHED
        """

        # Initialized
        INIT = 0b001

        # Currently running
        RUNNING = 0b010

        # All tasks completed
        COMPLETED = 0b100

        # finish() called
        FINISHED = 0b101

    _done = State.COMPLETED & State.FINISHED

    _state_transitions = {
        State.INIT: {
            State.RUNNING,
        },
        State.RUNNING: {
            State.COMPLETED,
            State.FINISHED,
        },
    }

    init_time = -1
    start_time = 0

    main_name = "main"

    def __init__(self):
        self._name = f"Kernel-{self.__class__._index}"
        self.__class__._index += 1

        self._state = self.State.INIT

        # Simulation time
        self._time: int = self.init_time

        # Main task
        self._main: Task | None = None

        # Currently executing task
        self._task: Task | None = None
        self._task_index = 0

        # Forked Tasks
        self._forks = dict[Task, set[Sendable]]()

        # Model variables
        self._dirty_vars = set[Variable]()

    def _set_state(self, state: State):
        assert state in self._state_transitions[self._state]
        self._state = state

    def state(self) -> State:
        """Current simulation state."""
        return self._state

    def time(self) -> int:
        """Current simulation time."""
        return self._time

    @property
    def main(self) -> Task:
        """Parent task of all other tasks."""
        assert self._main is not None
        return self._main

    def task(self) -> Task:
        """Currently running task."""
        assert self._task is not None
        return self._task

    def done(self) -> bool:
        """Return True if the kernel is done.

        A kernel that is "done" either:

        * Exhaused all tasks (COMPLETED), or
        * Called ``finish`` (FINISHED)
        """
        return bool(self._state & self._done)

    def clear(self):
        """Clear all simulation state."""
        self._forks.clear()
        self._dirty_vars.clear()

    # Scheduling methods
    @abstractmethod
    def call_soon(self, task: Task, args: TaskArgs) -> None:
        """Schedule task to run soon, in current time slot."""

    @abstractmethod
    def call_later(self, delay: int, task: Task, args: TaskArgs) -> None:
        """Schedule task to run later, after ``delay``."""

    @abstractmethod
    def call_at(self, when: int, task: Task, args: TaskArgs) -> None:
        """Schedule task to run at specified time: ``when``."""

    def _create_main(self, coro: TaskCoro) -> Task:
        assert self._time == self.init_time
        self._main = Task(coro, self.main_name)
        return self._main

    @abstractmethod
    def create_main(self, coro: TaskCoro) -> Task:
        """Create main task, and schedule it at time zero.

        Returns:
            Handle to the main task
        """

    def _create_task(self, coro: TaskCoro, name: str | None = None) -> Task:
        assert self._time >= self.start_time
        if name is None:
            name = f"Task-{self._task_index}"
            self._task_index += 1
        return Task(coro, name)

    @abstractmethod
    def create_task(self, coro: TaskCoro, name: str | None = None, **kwargs: Any) -> Task:
        """Create child task, and schedule it soon.

        Returns:
            Handle to the created task
        """

    def fork(self, task: Task, *ss: Sendable):
        self._forks[task] = set(ss)

    def join_any(self, task: Task, s: Sendable):
        if task in self._forks:
            ss = self._forks[task]
            ss.remove(s)
            while ss:
                s = ss.pop()
                s.wait_drop(task)
            del self._forks[task]

    def touch_var(self, v: Variable):
        self._dirty_vars.add(v)

    def _update_vars(self):
        while self._dirty_vars:
            v = self._dirty_vars.pop()
            v.update()

    def _start(self):
        if self._state is self.State.INIT:
            self._set_state(self.State.RUNNING)
        elif self._state is not self.State.RUNNING:
            s = f"Kernel has invalid state: {self._state.name}"
            raise RuntimeError(s)

    def _finish(self):
        self._set_state(self.State.FINISHED)

    @abstractmethod
    def _call(self, limit: int | None) -> None:
        """Run a simulation.

        Invoked by the public ``__call__`` method.
        Implements the inner loop of the top-level ``run`` function.
        """

    def __call__(self, ticks: int | None = None, until: int | None = None):
        # Determine the run limit
        if ticks is None:
            # Run until absolute limit, or no tasks left
            limit = until if until is not None else None
        else:
            # Run until relative limit
            limit = max(self.start_time, self._time) + ticks
            if until is not None:
                # Both relative & absolute given; clamp to soonest
                limit = min(limit, until)

        self._call(limit)

    @abstractmethod
    def _iter(self) -> Generator[int, None, None]:
        """Step (iterate) a simulation.

        Invoked by the public ``__iter__`` method.
        Implements the inner loop of the top-level ``step`` function.
        """

    def __iter__(self) -> Generator[int, None, None]:
        yield from self._iter()


class _PendQ(TaskQueue):
    """Priority queue for ordering task execution."""

    def __init__(self):
        # time, priority, index, task, value
        self._items = list[tuple[int, int, int, Task, Any]]()

        # Monotonically increasing integer
        # Breaks (time, priority, ...) ties in the heapq
        self._index: int = 0

    def __bool__(self) -> bool:
        return bool(self._items)

    def push(self, item: tuple[int, int, Task, Any]):
        time, priority, task, value = item
        task.link(self)
        heapq.heappush(self._items, (time, priority, self._index, task, value))
        self._index += 1

    def pop(self) -> tuple[Task, Any]:
        _, _, _, task, value = heapq.heappop(self._items)
        task.unlink(self)
        return (task, value)

    def _find(self, task: Task) -> int:
        for i, (_, _, _, t, _) in enumerate(self._items):
            if t is task:
                return i
        assert False  # pragma: no cover

    def drop(self, task: Task):
        index = self._find(task)
        self._items.pop(index)
        task.unlink(self)

    def peek(self) -> int:
        return self._items[0][0]

    def clear(self):
        while self._items:
            self.pop()
        self._index = 0


class DefaultKernel(Kernel):
    """Default simulation kernel

    Tasks are scheduled with a (heapq) priority queue.

    Task ordering rules:

    * Tasks scheduled at different times run in time order.
    * Tasks scheduled at same time run in priority order.
    * Tasks scheduled at same time with same priority run in insertion order.

    Priority is an arbitrary integer.

    The ``main`` (parent) task will be assigned priority zero.
    """

    main_priority = 0
    task_priority = 0

    def __init__(self):
        super().__init__()

        # Task queue
        self._queue = _PendQ()

        # Task priorities
        self._priorities = WeakKeyDictionary[Task, int]()

    @override
    def clear(self):
        super().clear()
        self._queue.clear()

    def call_soon(self, task: Task, args: TaskArgs):
        priority = self._priorities[task]
        self._queue.push((self._time, priority, task, args))

    def call_later(self, delay: int, task: Task, args: TaskArgs):
        priority = self._priorities[task]
        self._queue.push((self._time + delay, priority, task, args))

    def call_at(self, when: int, task: Task, args: TaskArgs):
        priority = self._priorities[task]
        self._queue.push((when, priority, task, args))

    def create_main(self, coro: TaskCoro) -> Task:
        main = super()._create_main(coro)
        self._priorities[main] = self.main_priority
        self.call_at(self.start_time, main, args=(Task.Command.START,))
        return main

    def create_task(self, coro: TaskCoro, name: str | None = None, **kwargs: Any) -> Task:
        task = super()._create_task(coro, name)
        self._priorities[task] = kwargs.get("priority", self.task_priority)
        self.call_soon(task, args=(Task.Command.START,))
        return task

    def _iter_time_slot(self, time: int) -> Generator[tuple[Task, TaskArgs], None, None]:
        """Iterate through all tasks in a time slot.

        The first task has already been peeked.
        This is a do-while loop.
        """
        task, value = self._queue.pop()
        yield (task, value)
        while self._queue and self._queue.peek() == time:
            task, value = self._queue.pop()
            yield (task, value)

    def _call(self, limit: int | None):
        self._start()

        while self._queue:
            # Peek when next event is scheduled
            time = self._queue.peek()

            # Protect against time traveling tasks
            assert time > self._time

            # Halt if we hit the run limit
            if limit is not None and time >= limit:
                return

            # Otherwise, advance to new timeslot
            self._time = time

            # Execute time slot
            for task, args in self._iter_time_slot(time):
                self._task = task
                try:
                    task.do_run(args)
                except _Finish:
                    self._finish()
                    return
                except StopIteration as exc:
                    task.do_result(exc)
                except Exception as exc:
                    task.do_except(exc)

            # Update simulation state
            self._update_vars()

        # Drop reference to last task
        self._task = None

        # All tasks exhausted
        self._set_state(self.State.COMPLETED)

    def _iter(self) -> Generator[int, None, None]:
        self._start()

        while self._queue:
            # Peek when next event is scheduled
            time = self._queue.peek()

            # Protect against time traveling tasks
            assert time > self._time

            # Yield before entering new timeslot
            yield time

            # Advance to new timeslot
            self._time = time

            # Execute time slot
            for task, args in self._iter_time_slot(time):
                self._task = task
                try:
                    task.do_run(args)
                except _Finish:
                    self._finish()
                    return
                except StopIteration as exc:
                    task.do_result(exc)
                except Exception as exc:
                    task.do_except(exc)

            # Update simulation state
            self._update_vars()

        # Drop reference to last task
        self._task = None

        # All tasks exhausted
        self._set_state(self.State.COMPLETED)


def finish():
    """Halt all incomplete coroutines, and immediately exit simulation.

    Clear all kernel data, and transition state to FINISHED.
    """
    raise _Finish()



"""Test dining philosophers."""

import random
from enum import Enum, auto

from pytest import CaptureFixture

from deltacycle import Lock, create_task, now, run, sleep

# Number of philosophers
N = 5


# Philosopher state
class State(Enum):
    INITIAL = auto()
    THINKING = auto()
    HUNGRY = auto()
    EATING = auto()


# Eat [min, max] time
EAT_TICKS = (50, 100)

# Think [min, max] time
THINK_TICKS = (50, 100)

# Simulation time
T = 1000


# Optional mask to filter print output
# If pmask & (1<<i), print philosopher state updates.
_pmask = (1 << N) - 1


# Philosophers and Forks
state = [State.INITIAL for _ in range(N)]
forks = [Lock() for _ in range(N)]


def init(pmask: int = (1 << N) - 1):
    """Initialize all philosophers and forks."""
    global _pmask, state, forks
    _pmask = pmask
    state = [State.INITIAL for _ in range(N)]
    forks = [Lock() for _ in range(N)]


def _update(i: int, ns: State):
    """Update philosopher[i] state."""
    if _pmask & (1 << i):
        print(f"[{now():08}] P{i} {state[i].name:8} => {ns.name:8}")
    state[i] = ns


async def think(i: int):
    """Philosopher[i] thinks for a random amount of time."""
    _update(i, State.THINKING)
    await sleep(random.randint(*THINK_TICKS))


async def pick_up_forks(i: int):
    """Philosopher[i] is hungry. Pick up left/right forks."""
    _update(i, State.HUNGRY)

    # Wait on forks in (left, right) order
    first, second = i, (i + 1) % N

    while True:
        # Wait until first fork is available
        await forks[first].get()

        # If second fork is available, get it.
        if forks[second].try_get():
            break

        # Second fork is NOT available:
        # 1. Release the first fork
        forks[first].put()
        # 2. Swap which fork we're waiting on first
        first, second = second, first


async def eat(i: int):
    """Philosopher[i] eats for a random amount of time."""
    _update(i, State.EATING)
    await sleep(random.randint(*EAT_TICKS))


def put_down_forks(i: int):
    """Philosopher[i] is not hungry. Put down left/right forks."""
    first, second = i, (i + 1) % N
    forks[first].put()
    forks[second].put()


async def philosopher(i: int):
    while True:
        await think(i)
        await pick_up_forks(i)
        await eat(i)
        put_down_forks(i)


async def main():
    for i in range(N):
        create_task(philosopher(i))


OUTPUT = """\
[00000000] P0 INITIAL  => THINKING
[00000000] P1 INITIAL  => THINKING
[00000000] P2 INITIAL  => THINKING
[00000000] P3 INITIAL  => THINKING
[00000000] P4 INITIAL  => THINKING
[00000051] P2 THINKING => HUNGRY
[00000051] P2 HUNGRY   => EATING
[00000057] P1 THINKING => HUNGRY
[00000067] P4 THINKING => HUNGRY
[00000067] P4 HUNGRY   => EATING
[00000090] P0 THINKING => HUNGRY
[00000097] P3 THINKING => HUNGRY
[00000116] P2 EATING   => THINKING
[00000116] P1 HUNGRY   => EATING
[00000131] P4 EATING   => THINKING
[00000131] P3 HUNGRY   => EATING
[00000174] P2 THINKING => HUNGRY
[00000187] P4 THINKING => HUNGRY
[00000213] P1 EATING   => THINKING
[00000213] P0 HUNGRY   => EATING
[00000224] P3 EATING   => THINKING
[00000224] P2 HUNGRY   => EATING
[00000279] P3 THINKING => HUNGRY
[00000297] P0 EATING   => THINKING
[00000297] P4 HUNGRY   => EATING
[00000310] P1 THINKING => HUNGRY
[00000311] P2 EATING   => THINKING
[00000311] P1 HUNGRY   => EATING
[00000349] P4 EATING   => THINKING
[00000349] P3 HUNGRY   => EATING
[00000362] P2 THINKING => HUNGRY
[00000366] P1 EATING   => THINKING
[00000374] P0 THINKING => HUNGRY
[00000374] P0 HUNGRY   => EATING
[00000412] P4 THINKING => HUNGRY
[00000413] P3 EATING   => THINKING
[00000413] P2 HUNGRY   => EATING
[00000448] P1 THINKING => HUNGRY
[00000462] P0 EATING   => THINKING
[00000462] P4 HUNGRY   => EATING
[00000464] P3 THINKING => HUNGRY
[00000498] P2 EATING   => THINKING
[00000498] P1 HUNGRY   => EATING
[00000524] P0 THINKING => HUNGRY
[00000557] P4 EATING   => THINKING
[00000557] P3 HUNGRY   => EATING
[00000589] P2 THINKING => HUNGRY
[00000592] P1 EATING   => THINKING
[00000592] P0 HUNGRY   => EATING
[00000633] P3 EATING   => THINKING
[00000633] P2 HUNGRY   => EATING
[00000641] P4 THINKING => HUNGRY
[00000656] P1 THINKING => HUNGRY
[00000670] P0 EATING   => THINKING
[00000670] P4 HUNGRY   => EATING
[00000700] P2 EATING   => THINKING
[00000700] P1 HUNGRY   => EATING
[00000720] P3 THINKING => HUNGRY
[00000720] P0 THINKING => HUNGRY
[00000760] P2 THINKING => HUNGRY
[00000768] P4 EATING   => THINKING
[00000768] P3 HUNGRY   => EATING
[00000794] P1 EATING   => THINKING
[00000794] P0 HUNGRY   => EATING
[00000839] P3 EATING   => THINKING
[00000839] P2 HUNGRY   => EATING
[00000845] P4 THINKING => HUNGRY
[00000853] P0 EATING   => THINKING
[00000853] P4 HUNGRY   => EATING
[00000861] P1 THINKING => HUNGRY
[00000902] P3 THINKING => HUNGRY
[00000909] P4 EATING   => THINKING
[00000924] P0 THINKING => HUNGRY
[00000924] P0 HUNGRY   => EATING
[00000937] P2 EATING   => THINKING
[00000937] P3 HUNGRY   => EATING
[00000964] P4 THINKING => HUNGRY
[00000993] P2 THINKING => HUNGRY
[00000998] P0 EATING   => THINKING
[00000998] P1 HUNGRY   => EATING
"""


def test_dp(capsys: CaptureFixture[str]):
    """This is a random algorithm, so we're only doing some basic checks."""
    random.seed(42)
    init()
    run(main(), until=T)
    out, _ = capsys.readouterr()
    out = "\n".join(line.rstrip() for line in out.splitlines()) + "\n"
    assert out == OUTPUT



"""Test deltacycle.CreditPool"""

# pyright: reportAttributeAccessIssue=false
# pyright: reportUnknownMemberType=false
# pyright: reportUnknownVariableType=false

import logging

import pytest
from pytest import LogCaptureFixture

from deltacycle import AllOf, CreditPool, TaskGroup, all_of, create_task, now, run, sleep

logger = logging.getLogger("deltacycle")


def test_len():
    async def main():
        credits = CreditPool(capacity=10)
        assert len(credits) == 0
        credits.put(1)
        assert len(credits) == 1
        credits.put(2)
        assert len(credits) == 3
        assert credits
        await credits.get(1)
        assert len(credits) == 2
        await credits.get(2)
        assert len(credits) == 0
        assert not credits

    run(main())


async def use_get_put(credits: CreditPool, t1: int, t2: int):
    logger.info("enter")

    await sleep(t1)

    logger.info("attempt get")
    await credits.get(n=2)
    logger.info("acquired")

    try:
        await sleep(t2)
    finally:
        logger.info("put")
        credits.put(n=2)

    await sleep(10)
    logger.info("exit")


async def use_try_get_put(credits: CreditPool, t1: int, t2: int):
    logger.info("enter")

    await sleep(t1)

    logger.info("attempt get")
    while not credits.try_get(n=2):
        await sleep(1)
    logger.info("acquired")

    try:
        await sleep(t2)
    finally:
        logger.info("put")
        credits.put(n=2)

    await sleep(10)
    logger.info("exit")


async def use_with(credits: CreditPool, t1: int, t2: int):
    logger.info("enter")

    await sleep(t1)

    logger.info("attempt get")
    async with credits.req(n=2):
        logger.info("acquired")
        await sleep(t2)
    logger.info("put")

    await sleep(10)
    logger.info("exit")


EXP1 = {
    # 0
    (0, "0", "enter"),
    (10, "0", "attempt get"),
    (10, "0", "acquired"),
    (20, "0", "put"),
    (30, "0", "exit"),
    # 1
    (0, "1", "enter"),
    (11, "1", "attempt get"),
    (11, "1", "acquired"),
    (21, "1", "put"),
    (31, "1", "exit"),
    # 2
    (0, "2", "enter"),
    (12, "2", "attempt get"),
    (12, "2", "acquired"),
    (22, "2", "put"),
    (32, "2", "exit"),
    # 3
    (0, "3", "enter"),
    (13, "3", "attempt get"),
    (13, "3", "acquired"),
    (23, "3", "put"),
    (33, "3", "exit"),
    # 4
    (0, "4", "enter"),
    (14, "4", "attempt get"),
    (14, "4", "acquired"),
    (24, "4", "put"),
    (34, "4", "exit"),
    # 5
    (0, "5", "enter"),
    (15, "5", "attempt get"),
    (20, "5", "acquired"),
    (30, "5", "put"),
    (40, "5", "exit"),
    # 6
    (0, "6", "enter"),
    (16, "6", "attempt get"),
    (21, "6", "acquired"),
    (31, "6", "put"),
    (41, "6", "exit"),
    # 7
    (0, "7", "enter"),
    (17, "7", "attempt get"),
    (22, "7", "acquired"),
    (32, "7", "put"),
    (42, "7", "exit"),
    # 8
    (0, "8", "enter"),
    (18, "8", "attempt get"),
    (23, "8", "acquired"),
    (33, "8", "put"),
    (43, "8", "exit"),
    # 9
    (0, "9", "enter"),
    (19, "9", "attempt get"),
    (24, "9", "acquired"),
    (34, "9", "put"),
    (44, "9", "exit"),
}

EXP2 = {
    # 0
    (0, "0", "enter"),
    (10, "0", "attempt get"),
    (10, "0", "acquired"),
    (20, "0", "put"),
    (30, "0", "exit"),
    # 1
    (0, "1", "enter"),
    (11, "1", "attempt get"),
    (11, "1", "acquired"),
    (21, "1", "put"),
    (31, "1", "exit"),
    # 2
    (0, "2", "enter"),
    (12, "2", "attempt get"),
    (12, "2", "acquired"),
    (22, "2", "put"),
    (32, "2", "exit"),
    # 3
    (0, "3", "enter"),
    (13, "3", "attempt get"),
    (13, "3", "acquired"),
    (23, "3", "put"),
    (33, "3", "exit"),
    # 4
    (0, "4", "enter"),
    (14, "4", "attempt get"),
    (14, "4", "acquired"),
    (24, "4", "put"),
    (34, "4", "exit"),
    # 5
    (0, "5", "enter"),
    (15, "5", "attempt get"),
    # NOTE: Ordering mismatch starts here.
    (24, "5", "acquired"),
    (34, "5", "put"),
    # NOTE: But total time (44) remains the same
    (44, "5", "exit"),
    # 6
    (0, "6", "enter"),
    (16, "6", "attempt get"),
    (23, "6", "acquired"),
    (33, "6", "put"),
    (43, "6", "exit"),
    # 7
    (0, "7", "enter"),
    (17, "7", "attempt get"),
    (22, "7", "acquired"),
    (32, "7", "put"),
    (42, "7", "exit"),
    # 8
    (0, "8", "enter"),
    (18, "8", "attempt get"),
    (21, "8", "acquired"),
    (31, "8", "put"),
    (41, "8", "exit"),
    # 9
    (0, "9", "enter"),
    (19, "9", "attempt get"),
    (20, "9", "acquired"),
    (30, "9", "put"),
    (40, "9", "exit"),
}


def test_get_put(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    async def main():
        credits = CreditPool(10)
        for i in range(10):
            create_task(use_get_put(credits, i + 10, 10), name=f"{i}")

    run(main())

    msgs = {(r.time, r.taskName, r.getMessage()) for r in caplog.records}
    assert msgs == EXP1


def test_try_get_put(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    async def main():
        credits = CreditPool(10)
        for i in range(10):
            create_task(use_try_get_put(credits, i + 10, 10), name=f"{i}")

    run(main())

    msgs = {(r.time, r.taskName, r.getMessage()) for r in caplog.records}
    assert msgs == EXP2


def test_async_with(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    async def main():
        credits = CreditPool(10)
        for i in range(10):
            create_task(use_with(credits, i + 10, 10), name=f"{i}")

    run(main())

    msgs = {(r.time, r.taskName, r.getMessage()) for r in caplog.records}
    assert msgs == EXP1


def test_bounded():
    async def use_bounded():
        credits = CreditPool(value=5, capacity=5)

        await credits.get(n=2)
        await credits.get(n=3)

        credits.put(n=2)
        credits.put(n=3)

        # Exception!
        with pytest.raises(OverflowError):
            credits.put(n=1)

    async def main():
        create_task(use_bounded())

    run(main())


def test_priority():
    async def request(credits: CreditPool, p: int):
        await sleep(5)
        await credits.get(n=2, priority=p)
        await sleep(5)
        credits.put()

    async def main():
        credits = CreditPool(value=2, capacity=2)
        t3 = create_task(request(credits, 3), name="T3")
        t2 = create_task(request(credits, 2), name="T2")
        t1 = create_task(request(credits, 1), name="T1")
        t0 = create_task(request(credits, 0), name="T0", priority=-1)

        ts = await all_of(t3, t2, t1, t0)
        assert ts == (t0, t1, t2, t3)
        assert now() == 25

    run(main())


def test_init_bad_values():
    with pytest.raises(ValueError):
        _ = CreditPool(value=5, capacity=4)

    with pytest.raises(ValueError):
        _ = CreditPool(value=-1)


def test_schedule_all1():
    async def cf(credits: CreditPool, t1: int, t2: int, t3: int):
        await sleep(t1)
        await credits.get(n=10)
        await sleep(t2)
        credits.put(n=10)
        await sleep(t3)

    async def main():
        credits = CreditPool(value=10, capacity=10)
        t1 = create_task(cf(credits, 0, 10, 10))

        await sleep(1)
        ys = await AllOf(t1, credits.req(n=10))

        assert ys == (credits, t1)
        assert now() == 20
        assert not credits
        credits.put(n=2)

    run(main())


def test_schedule_all2():
    async def cf(credits: CreditPool, t1: int, t2: int, t3: int):
        await sleep(t1)
        await credits.get(n=10)
        await sleep(t2)
        credits.put(n=10)
        await sleep(t3)

    async def main():
        credits = CreditPool(value=10, capacity=10)
        t1 = create_task(cf(credits, 0, 10, 10))

        await sleep(1)
        ys = await all_of(t1, credits.req(n=10))

        assert ys == (credits, t1)
        assert now() == 20
        assert not credits
        credits.put(n=10)

    run(main())


def test_overflow1():
    async def a(credits: CreditPool):
        credits.put(5)
        await sleep(5)

        # b gets 10
        await sleep(5)

        # 5 + 7 > 10 should overflow, despite the -10 from b
        credits.put(7)

    async def b(credits: CreditPool):
        # a puts 5
        await sleep(5)

        await credits.get(10)
        await sleep(5)

    async def main():
        credits = CreditPool(capacity=10)

        with pytest.raises(ExceptionGroup) as e:
            async with TaskGroup() as tg:
                tg.create_task(a(credits))
                tg.create_task(b(credits))

        excs = e.value.args[1]
        assert isinstance(excs[0], OverflowError)

    run(main())


def test_put_get_value_errors():
    async def main():
        credits = CreditPool(value=42, capacity=42)
        with pytest.raises(ValueError):
            credits.req(0)
        with pytest.raises(ValueError):
            credits.req(43)
        with pytest.raises(ValueError):
            credits.put(0)
        with pytest.raises(ValueError):
            credits.put(43)
        with pytest.raises(ValueError):
            credits.try_get(0)
        with pytest.raises(ValueError):
            credits.try_get(43)
        with pytest.raises(ValueError):
            await credits.get(0)
        with pytest.raises(ValueError):
            await credits.get(43)

    run(main())



"""Top-level functions."""

from collections import deque
from collections.abc import Generator
from typing import Any

from ._kernel import DefaultKernel, Kernel
from ._task import Blocking, Sendable, Task, TaskCoro

_kernel: Kernel | None = None


def get_running_kernel() -> Kernel:
    """Return currently running kernel.

    May be used by a simulation task to access kernel state.

    Returns:
        Kernel instance.

    Raises:
        RuntimeError: No kernel, or kernel is not currently running.
    """
    if _kernel is None:
        raise RuntimeError("No kernel")
    if _kernel.state() is not Kernel.State.RUNNING:
        raise RuntimeError("Kernel not RUNNING")
    return _kernel


def get_kernel() -> Kernel | None:
    """Get the current kernel.

    DeltaCycle only supports one simulation kernel at a time.
    This function gets the handle to that kernel, which may be ``None``.

    May be used by high level code to manage multiple kernels.

    Returns:
        Kernel instance or ``None``.
    """
    return _kernel


def set_kernel(kernel: Kernel | None = None):
    """Set the current kernel.

    DeltaCycle only supports one simulation kernel at a time.
    This function sets the handle to that kernel, which may be ``None``.

    May be used by high level code to manage multiple kernels.

    Args:
        kernel: Kernel instance or ``None``.
    """
    global _kernel  # noqa: PLW0603
    _kernel = kernel


def _get_kt() -> tuple[Kernel, Task]:
    kernel = get_running_kernel()
    task = kernel.task()
    return kernel, task


def get_current_task() -> Task:
    """Return currently running task.

    Returns:
        Task instance.

    Raises:
        RuntimeError: No kernel, or kernel is not currently running.
    """
    _, task = _get_kt()
    return task


def create_task(coro: TaskCoro, name: str | None = None, **kwargs: Any) -> Task:
    """Create a task, and schedule it to start soon.

    Args:
        coro: Coroutine function instance.
        name: Identify the task for logging/debugging purposes.
            If not given, a default name like ``Task-{index}`` will be assigned.
            Not guaranteed to be unique.
        kwargs: Arguments passed to the kernel to customize task execution.

    Returns:
        Created Task instance.

    Raises:
        RuntimeError: No kernel, or kernel is not currently running.
    """
    kernel = get_running_kernel()
    return kernel.create_task(coro, name, **kwargs)


def now() -> int:
    """Return current simulation time.

    Returns:
        Current simulation time.

    Raises:
        RuntimeError: No kernel, or kernel is not currently running.
    """
    kernel = get_running_kernel()
    return kernel.time()


def _run_pre(
    coro: TaskCoro | None,
    kernel: Kernel | None,
    kernel_type: type[Kernel],
) -> Kernel:
    if kernel is None:
        kernel = DefaultKernel()
        set_kernel(kernel)
        if coro is None:
            raise ValueError("New kernel requires a valid coro arg")
        assert coro is not None
        kernel.create_main(coro)
    else:
        set_kernel(kernel)

    return kernel


def run(
    coro: TaskCoro | None = None,
    kernel: Kernel | None = None,
    kernel_type: type[Kernel] = DefaultKernel,
    ticks: int | None = None,
    until: int | None = None,
) -> Any:
    """Run a simulation.

    If a simulation hits the run limit, it will exit and return ``None``.
    That simulation may be resumed any number of times.
    If all tasks are exhausted, return the main coroutine result.

    Args:
        coro: Main coroutine function instance.
            Required if creating a new kernel.
            Ignored if using an existing kernel.
        kernel: Optional Kernel instance.
            If not provided, a new kernel will be created.
        ticks: Optional relative run limit.
            If provided, run for *ticks* simulation time steps.
        until: Optional absolute run limit.
            If provided, run until *ticks* simulation time steps.

    Returns:
        If the main coroutine runs til completion, return its result.
        Otherwise, return ``None``.

    Raises:
        ValueError: Creating a new kernel, but no main coroutine provided.
        RuntimeError: The kernel is in an invalid state.
    """
    kernel = _run_pre(coro, kernel, kernel_type)
    kernel(ticks, until)

    if kernel.main.done():
        return kernel.main.result()


def step(
    coro: TaskCoro | None = None,
    kernel: Kernel | None = None,
    kernel_type: type[Kernel] = DefaultKernel,
) -> Generator[int, None, Any]:
    """Step (iterate) a simulation.

    Iterated simulations do not have a run limit.
    It is the user's responsibility to break at the desired time.
    If all tasks are exhausted, return the main coroutine result.

    Args:
        coro: Main coroutine function instance.
            Required if creating a new kernel.
            Ignored if using an existing kernel.
        kernel: Optional Kernel instance.
            If not provided, a new kernel will be created.

    Yields:
        Time immediately *before* the next time slot executes.

    Returns:
        Main coroutine result.

    Raises:
        ValueError: Creating a new kernel, but no main coroutine provided.
        RuntimeError: The kernel is in an invalid state.
    """
    kernel = _run_pre(coro, kernel, kernel_type)
    yield from kernel

    assert kernel.main.done()
    return kernel.main.result()


async def sleep(delay: int):
    """Suspend the current task, and wake up after a delay."""
    kernel, task = _get_kt()
    kernel.call_later(delay, task, args=(Task.Command.RESUME,))
    y = await task.switch_coro()
    assert y is None


async def all_of(*bs: Blocking) -> tuple[Sendable, ...]:
    """Block forward progress until all items are unblocked.

    Args:
        bs: Sequence of blocking items.

    Returns:
        Return a tuple of items in unblocking order.
    """
    _, task = _get_kt()

    blocked = set[Sendable]()
    unblocked = deque[Sendable]()

    for b in bs:
        if b.try_block(task):
            blocked.add(b.s)
        else:
            unblocked.append(b.s)

    while blocked:
        s = await task.switch_coro()
        assert isinstance(s, Sendable)
        blocked.remove(s)
        unblocked.append(s)

    return tuple(unblocked)


async def any_of(*bs: Blocking) -> Sendable | None:
    """Block forward progress until at least one item is unblocked.

    Args:
        bs: Sequence of blocking items.

    Returns:
        If the input is empty, return ``None``.
        Otherwise, return the item that unblocked first.
    """
    if not bs:
        return None

    kernel, task = _get_kt()

    blocked = set[Sendable]()

    for b in bs:
        if b.try_block(task):
            blocked.add(b.s)
        else:
            while blocked:
                s = blocked.pop()
                s.wait_drop(task)
            return b.s

    kernel.fork(task, *blocked)
    s = await task.switch_coro()
    assert isinstance(s, Sendable)
    return s



"""Delta Cycle

Credit to David Beazley's "Build Your Own Async" tutorial for inspiration:
https://www.youtube.com/watch?v=Y4Gt3Xjd7G8
"""

from ._container import Container
from ._credit_pool import CreditPool, ReqCredit
from ._event import Event
from ._kernel import Kernel, finish
from ._queue import Queue
from ._semaphore import Lock, ReqSemaphore, Semaphore
from ._task import (
    AllOf,
    AnyOf,
    Blocking,
    Interrupt,
    Sendable,
    Signal,
    Task,
    TaskCoro,
    TaskGroup,
)
from ._top import (
    all_of,
    any_of,
    create_task,
    get_current_task,
    get_kernel,
    get_running_kernel,
    now,
    run,
    set_kernel,
    sleep,
    step,
)
from ._variable import (
    Aggregate,
    AggrItem,
    AggrValue,
    Predicate,
    PredVar,
    Singular,
    Value,
    Variable,
)

__all__ = [
    # kernel
    "Kernel",
    "finish",
    "get_running_kernel",
    "get_kernel",
    "set_kernel",
    "run",
    "step",
    "now",
    "sleep",
    "all_of",
    "any_of",
    # container
    "Container",
    # event
    "Event",
    # queue
    "Queue",
    # credit_pool
    "CreditPool",
    "ReqCredit",
    # semaphore
    "Semaphore",
    "ReqSemaphore",
    "Lock",
    # task
    "TaskCoro",
    "Signal",
    "Interrupt",
    "Blocking",
    "Sendable",
    "AnyOf",
    "AllOf",
    "Task",
    "TaskGroup",
    "create_task",
    "get_current_task",
    # variable
    "Variable",
    "Predicate",
    "PredVar",
    "Value",
    "Singular",
    "Aggregate",
    "AggrItem",
    "AggrValue",
]



"""Semaphore synchronization primitive"""

from __future__ import annotations

from functools import cached_property
from types import TracebackType
from typing import Self

from ._kernel_if import KernelIf
from ._task import Blocking, SemaphoreQ, Sendable, Task


class Semaphore(KernelIf, Sendable):
    def __init__(self, value: int = 0, capacity: int = 0):
        self._capacity = capacity
        if value < 0:
            raise ValueError(f"Expected value ≥ 0, got {value}")
        if self._has_capacity and value > capacity:
            raise ValueError(f"Expected value ≤ {capacity}, got {value}")
        self._cnt = value
        self._waiting = SemaphoreQ()

    def __len__(self) -> int:
        return self._cnt

    @property
    def capacity(self) -> int | None:
        return self._capacity if self._has_capacity else None

    @cached_property
    def _has_capacity(self) -> bool:
        return self._capacity > 0

    def wait_push(self, priority: int, task: Task):
        self._waiting.push((priority, task))

    def wait_drop(self, task: Task):
        self._waiting.drop(task)

    def _check_cnt(self):
        assert self._cnt >= 0
        assert not self._has_capacity or self._cnt <= self._capacity

    def req(self, priority: int = 0) -> ReqSemaphore:
        return ReqSemaphore(self, priority)

    def put(self):
        self._check_cnt()

        if self._has_capacity and self._cnt == self._capacity:
            raise OverflowError(f"{self._cnt} + 1 > {self._capacity}")

        # Put credit
        self._cnt += 1

        if self._waiting:
            assert self._cnt == 1
            # Transfer credit
            task = self._waiting.pop()
            self._kernel.join_any(task, self)
            self._kernel.call_soon(task, args=(Task.Command.RESUME, self))
            self._cnt -= 1

    def try_get(self) -> bool:
        self._check_cnt()

        if self._cnt == 0:
            return False

        # Get credit
        self._cnt -= 1
        return True

    async def get(self, priority: int = 0):
        self._check_cnt()

        if self._cnt == 0:
            task = self._kernel.task()
            self.wait_push(priority, task)
            s = await task.switch_coro()
            assert s is self
        else:
            # Get credit
            self._cnt -= 1


class ReqSemaphore(Blocking):
    def __init__(self, sem: Semaphore, priority: int):
        self._sem = sem
        self._priority = priority

    async def __aenter__(self) -> Self:
        await self._sem.get(self._priority)
        return self

    async def __aexit__(
        self,
        exc_type: type[Exception],
        exc: Exception,
        traceback: TracebackType,
    ):
        self._sem.put()

    def try_block(self, task: Task) -> bool:
        if self._sem.try_get():
            return False

        self._sem.wait_push(self._priority, task)
        return True

    @property
    def s(self) -> Sendable:
        return self._sem


class Lock(Semaphore):
    def __init__(self):
        super().__init__(value=1, capacity=1)



"""Test bugs"""

from pytest import CaptureFixture

from deltacycle import TaskGroup, run, sleep

from .common import Bool, ttprint

EXP2 = """\
[   5][do_stuff] first
[  15][do_stuff] second
[  25][do_stuff] third
[  35][do_stuff] fourth
"""


def test_2(capsys: CaptureFixture[str]):
    clock = Bool(name="clock")

    async def do_stuff():
        await clock.posedge()
        ttprint("first")
        await clock.posedge()
        ttprint("second")
        await clock.posedge()
        ttprint("third")
        await clock.posedge()
        ttprint("fourth")

    async def drv_clock():
        clock.next = False
        while True:
            await sleep(5)
            clock.next = not clock.value

    async def main():
        async with TaskGroup() as tg:
            tg.create_task(drv_clock(), name="drv_clock")
            tg.create_task(do_stuff(), name="do_stuff")

    run(main(), until=100)

    cap = capsys.readouterr()
    # msgs = {(r.time, r.taskName, r.getMessage()) for r in caplog.records}
    # assert msgs == EXP2
    assert cap.out == EXP2



"""Queue synchronization primitive."""

from collections import deque
from functools import cached_property

from ._kernel_if import KernelIf
from ._task import SemaphoreQ, Task


class Queue[T](KernelIf):
    """Producer / Consumer FIFO Queue.

    Has both blocking and non-blocking put and get interfaces.
    If capacity is a positive number, the queue has *capacity* slots.
    If capacity is zero or a negative number, the queue has infinite slots.

    The put interface will block only when it is full.
    The get interface will block only when it is empty.

    An infinite queue will never be full.
    Its size is subject only to the machine's memory limitations.
    """

    def __init__(self, capacity: int = 0):
        self._capacity = capacity
        self._items = deque[T]()
        self._getq = SemaphoreQ()
        self._putq = SemaphoreQ()

    def __len__(self) -> int:
        return len(self._items)

    @property
    def capacity(self) -> int | None:
        return self._capacity if self._has_capacity else None

    @cached_property
    def _has_capacity(self) -> bool:
        return self._capacity > 0

    def empty(self) -> bool:
        return not self._items

    def full(self) -> bool:
        return self._has_capacity and len(self._items) == self._capacity

    def _put(self, item: T):
        self._items.append(item)
        if self._getq:
            task = self._getq.pop()
            self._kernel.call_soon(task, args=(Task.Command.RESUME,))

    def try_put(self, item: T) -> bool:
        """Nonblocking put: Return True if a put attempt is successful."""
        if self.full():
            return False

        self._put(item)
        return True

    async def put(self, item: T, priority: int = 0):
        """Block until there is space for an item."""
        if self.full():
            task = self._kernel.task()
            self._putq.push((priority, task))
            y = await task.switch_coro()
            assert y is None

        self._put(item)

    def _get(self) -> T:
        item = self._items.popleft()
        if self._putq:
            task = self._putq.pop()
            self._kernel.call_soon(task, args=(Task.Command.RESUME,))
        return item

    def try_get(self) -> tuple[bool, T | None]:
        """Nonblocking get.

        Returns:
            If the get is successful, ``(True, item)``;
            If unsuccessful, ``(False, None)``.
        """
        if self.empty():
            return False, None

        item = self._get()
        return True, item

    async def get(self, priority: int = 0) -> T:
        """Block until an item is available."""
        if self.empty():
            task = self._kernel.task()
            self._getq.push((priority, task))
            y = await task.switch_coro()
            assert y is None

        return self._get()



"""Test deltacycle.Task"""

# pyright: reportAttributeAccessIssue=false
# pyright: reportUnknownMemberType=false
# pyright: reportUnknownVariableType=false

import logging
from random import randint

import pytest
from pytest import LogCaptureFixture

from deltacycle import (
    AllOf,
    AnyOf,
    Event,
    Interrupt,
    Task,
    all_of,
    any_of,
    create_task,
    get_current_task,
    now,
    run,
    sleep,
    step,
)

logger = logging.getLogger("deltacycle")


def test_results():
    async def cf1(x0: int, x1: int) -> int:
        await sleep(randint(1, 5))
        return x0 * x1

    async def cf2(t0: Task, t1: Task) -> int:
        await t0
        await t1
        await sleep(randint(1, 5))
        return t0.result() * t1.result()

    async def main():
        t1 = create_task(cf1(2, 3))
        t2 = create_task(cf1(5, 7))
        t5 = create_task(cf2(t1, t2))

        t3 = create_task(cf1(11, 13))
        t4 = create_task(cf1(17, 19))
        t6 = create_task(cf2(t3, t4))

        t7 = create_task(c7 := cf2(t5, t6))

        assert t7.coro is c7

        r = await t7
        assert r == 2 * 3 * 5 * 7 * 11 * 13 * 17 * 19

    run(main())


def test_one_result():
    async def cf() -> int:
        await sleep(10)
        return 42

    async def main():
        t1 = create_task(cf())

        # Result/Exception not ready yet
        with pytest.raises(RuntimeError):
            t1.result()
        with pytest.raises(RuntimeError):
            t1.exception()

        r = await t1

        # Check Result/Exception
        assert r == 42
        assert t1.result() == 42
        assert t1.exception() is None

    run(main())


def test_one_exception():
    async def cf():
        await sleep(10)
        raise ArithmeticError(42)

    async def main():
        t1 = create_task(cf())

        # Result/Exception not ready yet
        with pytest.raises(RuntimeError):
            t1.result()
        with pytest.raises(RuntimeError):
            t1.exception()

        try:
            _ = await t1
        except ArithmeticError as e:
            # Check Result/Exception
            assert e.args[0] == 42
            try:
                _ = t1.result()
            except ArithmeticError as ee:
                assert ee is e
            assert t1.exception() is e
        else:
            assert False

    run(main())
    list(step(main()))


EXP1 = {
    # main
    (0, "main", "enter"),
    (5, "main", "interrupts C1"),
    (5, "main", "except"),
    (5, "main", "finally"),
    # C1
    (0, "C1", "enter"),
    (5, "C1", "except"),
    (5, "C1", "finally"),
    # C2
    (0, "C2", "enter"),
    (10, "C2", "finally"),
    # C3
    (0, "C3", "enter"),
    (10, "C3", "finally"),
}


def test_interrupt_pending1(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    async def cf(n: int):
        logger.info("enter")
        try:
            await sleep(n)
        except Interrupt:
            logger.info("except")
            raise
        finally:
            logger.info("finally")

    async def main():
        logger.info("enter")

        t1 = create_task(cf(1000), name="C1")
        t2 = create_task(cf(10), name="C2")
        t3 = create_task(cf(10), name="C3")

        await sleep(5)

        logger.info("interrupts C1")
        t1.interrupt()

        try:
            await t1
        except Interrupt as exc:
            logger.info("except")
            assert t1.exception() is exc
        finally:
            logger.info("finally")

        await t2
        await t3

        assert t1.done()
        assert t1.state() is Task.State.EXCEPTED

        # Result should re-raise Interrupt
        with pytest.raises(Interrupt):
            t1.result()

        # Cannot interrupt done task
        assert not t1.interrupt()

    run(main())
    msgs = {(r.time, r.taskName, r.getMessage()) for r in caplog.records}
    assert msgs == EXP1


def test_interrupt_pending2(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    async def cf(n: int):
        logger.info("enter")
        try:
            await sleep(n)
        except Interrupt:
            logger.info("except")
            raise
        finally:
            logger.info("finally")

    async def main():
        logger.info("enter")

        t1 = create_task(cf(1000), name="C1")
        t2 = create_task(cf(10), name="C2")
        t3 = create_task(cf(10), name="C3")

        await sleep(5)

        logger.info("interrupts C1")
        t1.interrupt()

        try:
            await t1
        except Interrupt as exc:
            logger.info("except")
            assert t1.exception() is exc
        finally:
            logger.info("finally")

        await t2
        await t3

        assert t1.done()
        assert t1.state() is Task.State.EXCEPTED

        # Result should re-raise Interrupt
        with pytest.raises(Interrupt):
            t1.result()

        # Cannot interrupt done task
        assert not t1.interrupt()

    list(step(main()))
    msgs = {(r.time, r.taskName, r.getMessage()) for r in caplog.records}
    assert msgs == EXP1


EXP2 = {
    # main
    (0, "main", "enter"),
    (20, "main", "exit"),
    # C1
    (0, "C1", "enter"),
    (10, "C1", "interrupted"),
    (10, "C1", "finally"),
    # C2
    (0, "C2", "enter"),
    (20, "C2", "finally"),
    # C3
    (0, "C3", "enter"),
    (20, "C3", "finally"),
}


def test_interrupt_waiting(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    async def cf(event: Event):
        logger.info("enter")
        try:
            await event
        except Interrupt:
            logger.info("interrupted")
        finally:
            logger.info("finally")

    async def main():
        logger.info("enter")
        event = Event()
        t1 = create_task(cf(event), name="C1")
        t2 = create_task(cf(event), name="C2")
        t3 = create_task(cf(event), name="C3")

        await sleep(10)

        t1.interrupt()

        await sleep(10)

        event.set()

        await t2
        await t3

        logger.info("exit")

    run(main())
    msgs = {(r.time, r.taskName, r.getMessage()) for r in caplog.records}
    assert msgs == EXP2


EXP3 = {
    (0, "Task started"),
    (1, "Main caught Interrupt from task"),
    (1, "Task interrupting itself"),
}


def test_interrupt_running(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    async def self_interrupting_task():
        logger.info("Task started")
        await sleep(1)
        task = get_current_task()
        logger.info("Task interrupting itself")
        task.interrupt()  # Interrupt self
        await sleep(1)  # This won't execute
        logger.info("This won't print")

    async def main():
        task = create_task(self_interrupting_task())
        try:
            await task
        except Interrupt:
            logger.info("Main caught Interrupt from task")

    run(main())
    msgs = {(r.time, r.getMessage()) for r in caplog.records}
    assert msgs == EXP3


def test_names():
    async def foo():
        await sleep(10)

    async def bar():
        fiz1 = create_task(fiz(), name="fiz")
        assert fiz1.name == "fiz"
        await sleep(10)

    async def fiz():
        await sleep(10)

    async def main():
        foo1 = create_task(foo())
        foo2 = create_task(foo())
        foo3 = create_task(foo(), name="foo")
        foo4 = create_task(foo(), name="foo")

        assert foo1.name.startswith("Task-")
        assert foo2.name.startswith("Task-")
        assert foo3.name == "foo"
        assert foo4.name == "foo"

        create_task(bar(), name="bar")

    run(main())


def test_task_any1():
    async def cf(t: int):
        await sleep(t)

    async def main():
        t1 = create_task(cf(5), name="T1")
        t2 = create_task(cf(10), name="T2")
        t3 = create_task(cf(15), name="T3")

        t = await any_of(t1, t2, t3)
        assert t is t1
        assert now() == 5

    run(main())


def test_task_any2():
    async def main():
        y = await AnyOf()
        assert y is None
        assert now() == 0

    run(main())


def test_task_all1():
    async def cf(t: int):
        await sleep(t)

    async def main():
        t1 = create_task(cf(5), name="T1")
        t2 = create_task(cf(10), name="T2")
        t3 = create_task(cf(15), name="T3")

        ts = await all_of(t1, t2, t3)
        assert ts == (t1, t2, t3)
        assert now() == 15

    run(main())


def test_task_all2():
    async def cf(t: int):
        await sleep(t)

    async def main():
        t1 = create_task(cf(5), name="T1")
        t2 = create_task(cf(10), name="T2")
        t3 = create_task(cf(15), name="T3")

        await sleep(6)  # t1 is done

        ts = await AllOf(t1, t2, t3)
        assert ts == (t1, t2, t3)
        assert now() == 15

    run(main())


def test_task_all3():
    async def cf(t: int):
        await sleep(t)

    async def main():
        t1 = create_task(cf(5), name="T1")
        t2 = create_task(cf(10), name="T2")
        t3 = create_task(cf(15), name="T3")

        await sleep(20)  # all are done

        ts = await all_of(t1, t2, t3)
        assert ts == (t1, t2, t3)
        assert now() == 20

    run(main())


def test_task_all4():
    async def cf(t: int):
        await sleep(t)

    async def main():
        t1 = create_task(cf(5), name="T1")
        t2 = create_task(cf(10), name="T2")
        t3 = create_task(cf(15), name="T3")

        ts = await AllOf(t1, t2, t3)
        assert ts == (t1, t2, t3)
        assert now() == 15

    run(main())



"""Test seqlogic.sim module."""

from pytest import CaptureFixture

from deltacycle import Kernel, create_task, get_running_kernel, run, sleep, step

from .common import Bool, tprint


async def drv_clk(clk: Bool):
    while True:
        await sleep(5)
        clk.next = not clk.prev


async def drv_a(a: Bool, clk: Bool):
    while True:
        await clk.edge()
        a.next = not a.prev


async def drv_b(b: Bool, clk: Bool):
    i = 0
    while True:
        await clk.edge()
        if i % 2 == 0:
            b.next = not b.prev
        else:
            b.next = b.prev
        i += 1


async def drv_c(c: Bool, clk: Bool):
    i = 0
    while True:
        await clk.edge()
        if i % 3 == 0:
            c.next = not c.prev
        i += 1


async def mon(a: Bool, b: Bool, c: Bool, clk: Bool):
    while True:
        await clk.edge()
        tprint(f"a={a.prev:b} b={b.prev:b} c={c.prev:b}")


EXP = """\
[   5] a=0 b=0 c=0
[  10] a=1 b=1 c=1
[  15] a=0 b=1 c=1
[  20] a=1 b=0 c=1
[  25] a=0 b=0 c=0
[  30] a=1 b=1 c=0
[  35] a=0 b=1 c=0
[  40] a=1 b=0 c=1
[  45] a=0 b=0 c=1
"""


def test_vars_run(capsys: CaptureFixture[str]):
    """Test run, halt, run."""

    clk = Bool(name="clk")
    a = Bool(name="a")
    b = Bool(name="b")
    c = Bool(name="c")

    async def main():
        create_task(drv_clk(clk), priority=2)
        create_task(drv_a(a, clk), priority=2)
        create_task(drv_b(b, clk), priority=2)
        create_task(drv_c(c, clk), priority=2)
        create_task(mon(a, b, c, clk), priority=3)

    # Relative run limit
    run(main(), ticks=25)

    kernel = get_running_kernel()

    # Absolute run limit
    run(kernel=kernel, until=50)

    cap = capsys.readouterr()
    assert cap.out == EXP


def test_vars_iter(capsys: CaptureFixture[str]):
    """Test iter, iter."""

    clk = Bool(name="clk")
    a = Bool(name="a")
    b = Bool(name="b")
    c = Bool(name="c")

    async def main():
        create_task(drv_clk(clk), priority=2)
        create_task(drv_a(a, clk), priority=2)
        create_task(drv_b(b, clk), priority=2)
        create_task(drv_c(c, clk), priority=2)
        create_task(mon(a, b, c, clk), priority=3)

    for t in step(main()):
        if t >= 25:
            break

    kernel = get_running_kernel()
    assert kernel.state() is Kernel.State.RUNNING

    for t in step(kernel=kernel):
        if t >= 50:
            break

    assert kernel.state() is Kernel.State.RUNNING

    cap = capsys.readouterr()
    assert cap.out == EXP


def test_vars_run_iter(capsys: CaptureFixture[str]):
    """Test run, halt, iter."""

    clk = Bool(name="clk")
    a = Bool(name="a")
    b = Bool(name="b")
    c = Bool(name="c")

    async def main():
        create_task(drv_clk(clk), priority=2)
        create_task(drv_a(a, clk), priority=2)
        create_task(drv_b(b, clk), priority=2)
        create_task(drv_c(c, clk), priority=2)
        create_task(mon(a, b, c, clk), priority=3)

    # Relative run limit
    run(main(), ticks=25)

    kernel = get_running_kernel()

    for t in step(kernel=kernel):
        if t >= 50:
            break

    assert kernel.state() is Kernel.State.RUNNING

    cap = capsys.readouterr()
    assert cap.out == EXP



"""Test Bank Renege Example

The following is a copy/rewrite of the "Bank Renege" example, by Team SimPy.
The purpose is to test similarities/differences with DeltaCycle.
See SimPy license for copyright details.

Reference:
* https://simpy.readthedocs.io/en/latest/about/license.html
* https://simpy.readthedocs.io/en/latest/examples/bank_renege.html
"""

import random

from pytest import CaptureFixture

from deltacycle import AnyOf, Lock, create_task, now, run, sleep

RANDOM_SEED = 42
NEW_CUSTOMERS = 5  # Number of customers
INTERVAL_CUSTOMERS = 10.0  # Generate new customers roughly every x seconds
MIN_PATIENCE = 1.0  # Min customer patience
MAX_PATIENCE = 3.0  # Max customer patience

TIME_IN_BANK = 12.0

TIMESCALE = 1_000_000.0


def tprint(s: str):
    print(f"{now() / TIMESCALE:7.4f} {s}")


async def customer(name: str, counter: Lock):
    """Customer arrives, is served and leaves."""
    arrive = now()
    tprint(f"{name}: Here I am")

    patience = random.uniform(MIN_PATIENCE, MAX_PATIENCE)

    # Wait for the counter or abort at the end of our tether
    timeout = create_task(sleep(round(patience * TIMESCALE)))
    y = await AnyOf(counter.req(), timeout)
    wait = now() - arrive

    if y is counter:
        # We got to the counter
        tprint(f"{name}: Waited {wait / TIMESCALE:<7.3f}")
        t = random.expovariate(1.0 / TIME_IN_BANK)
        await sleep(round(t * TIMESCALE))
        tprint(f"{name}: Finished")
        assert isinstance(y, Lock)
        y.put()
    else:
        # We reneged
        tprint(f"{name}: RENEGED after {wait / TIMESCALE:<7.3f}")


async def main(n: int, interval: float, counter: Lock):
    """Generate customers randomly."""
    for i in range(n):
        c = customer(f"Customer{i:02d}", counter)
        create_task(c)
        t = random.expovariate(1.0 / interval)
        await sleep(round(t * TIMESCALE))


OUTPUT = """\
Bank Renege
 0.0000 Customer00: Here I am
 0.0000 Customer00: Waited 0.000
 3.8595 Customer00: Finished
10.2006 Customer01: Here I am
10.2006 Customer01: Waited 0.000
12.7265 Customer02: Here I am
13.9003 Customer02: RENEGED after 1.174
23.7507 Customer01: Finished
34.9993 Customer03: Here I am
34.9993 Customer03: Waited 0.000
37.9599 Customer03: Finished
40.4798 Customer04: Here I am
40.4798 Customer04: Waited 0.000
43.1401 Customer04: Finished
"""


def test_bank_renege(capsys: CaptureFixture[str]):
    # Setup and start the simulation
    print("Bank Renege")
    random.seed(RANDOM_SEED)

    # Start processes and run
    counter = Lock()
    run(main(NEW_CUSTOMERS, INTERVAL_CUSTOMERS, counter))
    out, _ = capsys.readouterr()
    out = "\n".join(line.rstrip() for line in out.splitlines()) + "\n"
    assert out == OUTPUT



"""Hello, world!"""

import pytest
from pytest import CaptureFixture

from deltacycle import Kernel, get_kernel, run, sleep

from .common import tprint

EXP = """\
[  -1] Before Time
[   2] Hello
[   4] World
"""


def test_hello(capsys: CaptureFixture[str]):
    """Test basic async/await hello world functionality."""
    tprint("Before Time")

    async def hello():
        await sleep(2)
        tprint("Hello")
        await sleep(2)
        tprint("World")
        return 42

    ret = run(hello())
    assert ret == 42

    kernel = get_kernel()
    assert kernel is not None
    assert kernel.state() is Kernel.State.COMPLETED

    with pytest.raises(RuntimeError):
        run(kernel=kernel)

    cap = capsys.readouterr()
    assert cap.out == EXP



"""Event synchronization primitive"""

from __future__ import annotations

from collections.abc import Generator
from typing import Self

from ._kernel_if import KernelIf
from ._task import Blocking, EventQ, Sendable, Task


class Event(KernelIf, Blocking, Sendable):
    """Notify multiple tasks that some event has happened.

    An event instance is lightweight.
    It consists of a flag, and a FIFO of waiting tasks.

    When the event is created, its flag defaults to ``False``.
    In this state, the event will block all tasks that await it.
    When a task invokes the event's ``set`` method,
    that sets the flag (to ``True``), and unblocks all waiting tasks.

    If the event's flag is set, it will not block awaiting tasks.
    When a task invokes the event's ``clear`` method,
    that clears the flag (to ``False``),
    and the event will go back to blocking awaiting tasks.
    """

    def __init__(self):
        self._flag = False
        self._waiting = EventQ()

    def _blocking(self) -> bool:
        return not self._flag

    def wait_push(self, task: Task):
        self._waiting.push(task)

    def wait_drop(self, task: Task):
        self._waiting.drop(task)

    def __await__(self) -> Generator[None, Sendable, Self]:
        """Await event set."""
        if self._blocking():
            task = self._kernel.task()
            self.wait_push(task)
            e = yield from task.switch_gen()
            assert e is self

        return self

    def __bool__(self) -> bool:
        """Return flag state."""
        return self._flag

    def set(self):
        """Set the flag. Stop blocking waiting tasks."""
        self._flag = True
        self._waiting.load()

        while self._waiting:
            task = self._waiting.pop()
            self._kernel.join_any(task, self)
            self._kernel.call_soon(task, args=(Task.Command.RESUME, self))

    def clear(self):
        """Clear the flag. Start blocking waiting tasks."""
        self._flag = False

    # Blocking
    def try_block(self, task: Task) -> bool:
        if self._blocking():
            self.wait_push(task)
            return True
        return False

    @property
    def s(self) -> Event:
        return self



"""Test deltacycle.Container"""

# pyright: reportAttributeAccessIssue=false
# pyright: reportUnknownMemberType=false
# pyright: reportUnknownVariableType=false

import logging

import pytest
from pytest import LogCaptureFixture

from deltacycle import Container, TaskGroup, run, sleep

logger = logging.getLogger("deltacycle")


def test_len():
    async def main():
        container = Container(capacity=10)
        assert container.capacity == 10
        assert len(container) == 0
        await container.put(1)
        assert len(container) == 1
        await container.put(2)
        assert len(container) == 3
        assert container
        await container.get(1)
        assert len(container) == 2
        await container.get(2)
        assert len(container) == 0
        assert not container

    run(main())


EXP1 = {
    # P1
    (0, "P1: PUT"),
    (1, "P1: PUT"),
    (2, "P1: PUT"),
    (3, "P1: PUT"),
    (4, "P1: PUT"),
    (5, "P1: PUT"),
    (6, "P1: PUT"),
    (7, "P1: PUT"),
    (8, "P1: PUT"),
    (9, "P1: PUT"),
    (10, "P1: PUT"),
    (11, "P1: PUT"),
    (12, "P1: PUT"),
    (14, "P1: PUT"),
    (15, "P1: PUT"),
    (16, "P1: PUT"),
    # P2
    (0, "P2: PUT"),
    (1, "P2: PUT"),
    (3, "P2: PUT"),
    (4, "P2: PUT"),
    (5, "P2: PUT"),
    (7, "P2: PUT"),
    (8, "P2: PUT"),
    (10, "P2: PUT"),
    (11, "P2: PUT"),
    (12, "P2: PUT"),
    (13, "P2: PUT"),
    (14, "P2: PUT"),
    # C1
    (0, "C1: GET"),
    (1, "C1: GET"),
    (2, "C1: GET"),
    (3, "C1: GET"),
    (4, "C1: GET"),
    (5, "C1: GET"),
    (6, "C1: GET"),
    (7, "C1: GET"),
    (8, "C1: GET"),
    (9, "C1: GET"),
    (10, "C1: GET"),
    (11, "C1: GET"),
    (12, "C1: GET"),
    (13, "C1: GET"),
    (14, "C1: GET"),
    (15, "C1: GET"),
    # C2
    (0, "C2: GET"),
    (1, "C2: GET"),
    (2, "C2: GET"),
    (3, "C2: GET"),
    (4, "C2: GET"),
    (5, "C2: GET"),
    (6, "C2: GET"),
    (7, "C2: GET"),
    (8, "C2: GET"),
    (9, "C2: GET"),
    (10, "C2: GET"),
    (11, "C2: GET"),
    (12, "C2: GET"),
    (13, "C2: GET"),
    (14, "C2: GET"),
    (15, "C2: GET"),
    # C3
    (0, "C3: GET"),
    (1, "C3: GET"),
    (3, "C3: GET"),
    (4, "C3: GET"),
    (5, "C3: GET"),
    (6, "C3: GET"),
    (7, "C3: GET"),
    (8, "C3: GET"),
    (9, "C3: GET"),
    (10, "C3: GET"),
    (11, "C3: GET"),
    (12, "C3: GET"),
    (13, "C3: GET"),
    (14, "C3: GET"),
    (15, "C3: GET"),
    (16, "C3: GET"),
}


def test_prod_cons1(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    async def p(name: str, container: Container, n: int):
        for _ in range(16):
            await container.put(n)
            logger.info(f"{name}: PUT")
            await sleep(1)

    async def c(name: str, container: Container, n: int):
        for _ in range(16):
            await container.get(n)
            logger.info(f"{name}: GET")
            await sleep(1)

    async def main():
        container = Container(capacity=10)

        async with TaskGroup() as tg:
            tg.create_task(p("P1", container, 5))
            tg.create_task(p("P2", container, 7))
            tg.create_task(c("C1", container, 2))
            tg.create_task(c("C2", container, 3))
            tg.create_task(c("C3", container, 5))

    run(main())

    msgs = {(r.time, r.getMessage()) for r in caplog.records}
    assert msgs == EXP1


EXP2 = {
    # P1
    (0, "P1: PUT"),
    (1, "P1: PUT"),
    (2, "P1: PUT"),
    (3, "P1: PUT"),
    (4, "P1: PUT"),
    (5, "P1: PUT"),
    (6, "P1: PUT"),
    (7, "P1: PUT"),
    # P2
    (0, "P2: PUT"),
    (1, "P2: PUT"),
    (2, "P2: PUT"),
    (3, "P2: PUT"),
    (4, "P2: PUT"),
    (5, "P2: PUT"),
    (6, "P2: PUT"),
    # P3
    (8, "P3: PUT"),
    (9, "P3: PUT"),
    (10, "P3: PUT"),
    (11, "P3: PUT"),
    (12, "P3: PUT"),
    (13, "P3: PUT"),
    (14, "P3: PUT"),
    (15, "P3: PUT"),
    # C1
    (0, "C1: GET"),
    (1, "C1: GET"),
    (2, "C1: GET"),
    (3, "C1: GET"),
    (4, "C1: GET"),
    (5, "C1: GET"),
    (6, "C1: GET"),
    (7, "C1: GET"),
    # C2
    (0, "C2: GET"),
    (1, "C2: GET"),
    (2, "C2: GET"),
    (3, "C2: GET"),
    (4, "C2: GET"),
    (5, "C2: GET"),
    (6, "C2: GET"),
    (7, "C2: GET"),
    # C3
    (7, "P2: PUT"),
    (8, "C3: GET"),
    (9, "C3: GET"),
    (10, "C3: GET"),
    (11, "C3: GET"),
    (12, "C3: GET"),
    (13, "C3: GET"),
    (14, "C3: GET"),
    (15, "C3: GET"),
}


def test_prod_cons2(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    async def p(name: str, container: Container, n: int):
        for _ in range(8):
            while not container.try_put(n):
                await sleep(1)
            logger.info(f"{name}: PUT")
            await sleep(1)

    async def c(name: str, container: Container, n: int):
        for _ in range(8):
            while not container.try_get(n):
                await sleep(1)
            logger.info(f"{name}: GET")
            await sleep(1)

    async def main():
        container = Container(capacity=8)

        async with TaskGroup() as tg:
            tg.create_task(p("P1", container, 2))
            tg.create_task(p("P2", container, 3))
            tg.create_task(p("P3", container, 5))
            tg.create_task(c("C1", container, 2))
            tg.create_task(c("C2", container, 3))
            tg.create_task(c("C3", container, 5))

    run(main())

    msgs = {(r.time, r.getMessage()) for r in caplog.records}
    assert msgs == EXP2


def test_put_get_value_errors():
    async def main():
        container = Container(capacity=42)
        with pytest.raises(ValueError):
            await container.put(0)
        with pytest.raises(ValueError):
            await container.put(43)
        with pytest.raises(ValueError):
            container.try_get(0)
        with pytest.raises(ValueError):
            container.try_get(43)
        with pytest.raises(ValueError):
            await container.get(0)
        with pytest.raises(ValueError):
            await container.get(43)

    run(main())



"""Test deltacycle.Event"""

# pyright: reportAttributeAccessIssue=false
# pyright: reportPrivateUsage=false
# pyright: reportUnknownMemberType=false
# pyright: reportUnknownVariableType=false

import logging

from pytest import LogCaptureFixture

from deltacycle import AnyOf, Event, any_of, create_task, get_running_kernel, now, run, sleep

logger = logging.getLogger("deltacycle")


async def primary(event: Event):
    logger.info("enter")

    await sleep(10)

    # T=10
    logger.info("set")
    event.set()
    assert event

    await sleep(10)

    # T=20
    logger.info("clear")
    event.clear()
    assert not event

    await sleep(10)

    # T=30
    logger.info("set")
    event.set()
    assert event

    logger.info("exit")


async def secondary(event: Event):
    logger.info("enter")

    # Event clear
    logger.info("waiting")
    await event

    # Event set @10
    logger.info("running")
    await sleep(10)

    # Event clear
    logger.info("waiting")
    await event

    # Event set @30
    logger.info("running")
    await sleep(10)

    # Event still set: return immediately
    await event

    logger.info("exit")


EXP1 = {
    # P
    (0, "P", "enter"),
    (10, "P", "set"),
    (20, "P", "clear"),
    (30, "P", "set"),
    (30, "P", "exit"),
    # S1
    (0, "S1", "enter"),
    (0, "S1", "waiting"),
    (10, "S1", "running"),
    (20, "S1", "waiting"),
    (30, "S1", "running"),
    (40, "S1", "exit"),
    # S2
    (0, "S2", "enter"),
    (0, "S2", "waiting"),
    (10, "S2", "running"),
    (20, "S2", "waiting"),
    (30, "S2", "running"),
    (40, "S2", "exit"),
    # S3
    (0, "S3", "enter"),
    (0, "S3", "waiting"),
    (10, "S3", "running"),
    (20, "S3", "waiting"),
    (30, "S3", "running"),
    (40, "S3", "exit"),
}


def test_acquire_release(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    async def main():
        event = Event()
        create_task(primary(event), name="P")
        create_task(secondary(event), name="S1")
        create_task(secondary(event), name="S2")
        create_task(secondary(event), name="S3")

    run(main())

    msgs = {(r.time, r.taskName, r.getMessage()) for r in caplog.records}
    assert msgs == EXP1


def test_serial():
    async def sleep_set(e: Event, t: int):
        await sleep(t)
        e.set()

    async def main():
        e0 = Event()
        e1 = Event()
        e2 = Event()

        create_task(sleep_set(e0, 10), name="first")
        create_task(sleep_set(e1, 20), name="second")
        create_task(sleep_set(e2, 30), name="third")

        e: Event = await e0
        assert e is e0 and now() == 10
        e: Event = await e1
        assert e is e1 and now() == 20
        e: Event = await e2
        assert e is e2 and now() == 30

    run(main())


def test_list_any_list():
    async def sleep_set(t: int, e: Event):
        await sleep(t)
        e.set()

    async def main():
        e10 = Event()
        e20 = Event()
        e30 = Event()

        create_task(sleep_set(10, e10), name="e10")
        create_task(sleep_set(20, e20), name="e20")
        create_task(sleep_set(30, e30), name="e30")

        e = await any_of(e10, e20, e30)
        assert e is e10

        kernel = get_running_kernel()
        assert not kernel._forks

    run(main())


def test_list_any_one_set():
    async def sleep_set(t: int, e: Event):
        await sleep(t)
        e.set()

    async def main():
        e10 = Event()
        e20 = Event()
        e30 = Event()

        e20.set()

        create_task(sleep_set(10, e10), name="e10")
        create_task(sleep_set(20, e20), name="e20")
        create_task(sleep_set(30, e30), name="e30")

        e = await AnyOf(e10, e20, e30)
        assert e is e20
        assert now() == 0

        kernel = get_running_kernel()
        assert not kernel._forks

    run(main())



"""Model variables"""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections import OrderedDict, defaultdict, deque
from collections.abc import Callable, Generator, Hashable
from typing import Self

from ._kernel_if import KernelIf
from ._task import Blocking, Sendable, Task, TaskQueue

type Predicate = Callable[[], bool]


class _WaitQ(TaskQueue):
    """Tasks wait for variable touch."""

    def __init__(self):
        self._t2p = OrderedDict[Task, Predicate]()
        self._items = deque[Task]()

    def __bool__(self) -> bool:
        return bool(self._items)

    def push(self, item: tuple[Task, Predicate]):
        task, p = item
        task.link(self)
        self._t2p[task] = p

    def pop(self) -> Task:
        task = self._items.popleft()
        self.drop(task)
        return task

    def drop(self, task: Task):
        del self._t2p[task]
        task.unlink(self)

    def load(self):
        assert not self._items
        self._items.extend(t for t, p in self._t2p.items() if p())


class Variable(KernelIf, Blocking, Sendable):
    """Model component that changes over time.

    The instantaneous state of a simulation is represented by a collection of variables.

    There are two types of variables: *singular*, and *aggregate*.

    Children::

               Variable
                  |
           +------+------+
           |             |
        Singular     Aggregate

    * A singular variable has one *value*.
    * An aggregate variable is a mapping of key, *value* pairs.

    Variables are *always* blocking.
    Tasks may schedule updates to variables.
    Changes to variable values may unblock tasks,
    which may in turn schedule updates to other variables.
    """

    def __init__(self):
        self._waiting = _WaitQ()

    def wait_push(self, task: Task, p: Predicate):
        self._waiting.push((task, p))

    def wait_drop(self, task: Task):
        self._waiting.drop(task)

    def __await__(self) -> Generator[None, Sendable, Self]:
        """Await variable change:

        For variable ``v``:

        1. Suspend the current task.
        2. When another task invokes ``v.set_next(...)`` *and*
           ``v.changed()`` evaluates to ``True``,
           unblock all tasks waiting for that event.
        """
        task = self._kernel.task()
        # NOTE: Use default predicate
        self.wait_push(task, self.changed)
        v = yield from task.switch_gen()
        assert v is self
        return self

    def _set(self):
        self._waiting.load()

        while self._waiting:
            task = self._waiting.pop()
            self._kernel.join_any(task, self)
            self._kernel.call_soon(task, args=(Task.Command.RESUME, self))

        # Add variable to update set
        self._kernel.touch_var(self)

    def pred(self, p: Predicate) -> PredVar:
        """Return blocking, predicated variable.

        Args:
            p: Prediate function w/ no args and ``bool`` return type.

        Returns:
            Predicated Variable (``PredVar``) object.
        """
        return PredVar(self, p)

    @abstractmethod
    def changed(self) -> bool:
        """Return True if changed during the current time slot."""

    @abstractmethod
    def update(self) -> None:
        """Update variable value."""

    # Blocking
    def try_block(self, task: Task) -> bool:
        # NOTE: Use default predicate
        self.wait_push(task, self.changed)
        return True

    @property
    def s(self) -> Variable:
        return self


class PredVar(KernelIf, Blocking):
    """Predicated Variable.

    A lightweight wrapper around a Variable instance.
    Implements ``Awaitable`` and ``Blocking``.
    Can be used in ``await``, ``await AllOf`` and ``await AnyOf`` statements.

    Predicate functions allow fine-grained control of variable dependencies.
    Sometimes waiting tasks can be woken up when there is any change to the
    variables's value. However, it is often desirable to only trigger on
    particular types of changes. For example, in digital logic a flip-flop
    might only update its state when 1) reset is inactive, 2) clock is
    transitioning from low to high (a positive edge), AND 3) a data enable
    signal is active. A predicate function may be used to evaluate when
    those conditions are all true.
    """

    def __init__(self, var: Variable, p: Predicate):
        self._var = var
        self._p = p

    def __await__(self) -> Generator[None, Sendable, Variable]:
        """Await variable change:

        For variable ``v``, and predicate function ``p``:

        1. Suspend the current task.
        2. When another task invokes ``v.set_next(...)`` *and* ``p`` evaluates
           to ``True``, unblock all tasks waiting for that event.
        """
        task = self._kernel.task()
        self._var.wait_push(task, self._p)
        v = yield from task.switch_gen()
        assert v is self._var
        return self._var

    # Blocking
    def try_block(self, task: Task) -> bool:
        self._var.wait_push(task, self._p)
        return True

    @property
    def s(self) -> Sendable:
        return self._var


class Value[T](ABC):
    """Variable value."""

    @abstractmethod
    def get_prev(self) -> T:
        """Return value at the end of the previous timeslot."""

    prev = property(fget=get_prev)

    @abstractmethod
    def set_next(self, value: T) -> None:
        """Schedule update to value in the current timeslot."""

    next = property(fset=set_next)


class Singular[T](Variable, Value[T]):
    """Model state organized as a single unit."""

    def __init__(self, value: T):
        Variable.__init__(self)
        self._prev = value
        self._next = value
        self._changed: bool = False

    # Value
    def get_prev(self) -> T:
        return self._prev

    prev = property(fget=get_prev)

    def set_next(self, value: T):
        self._changed = value != self._next
        self._next = value

        # Notify the kernel
        self._set()

    next = property(fset=set_next)

    # Variable
    def get_value(self) -> T:
        """Return present value.

        When performing multiple updates to a variable during the same timeslot,
        this method will always return the value of the *latest* update.
        """
        return self._next

    value = property(fget=get_value)

    def changed(self) -> bool:
        return self._changed

    def update(self):
        self._prev = self._next
        self._changed = False


class Aggregate[T](Variable):
    """Model state organized as multiple units."""

    def __init__(self, value: T):
        Variable.__init__(self)
        self._prevs = defaultdict[Hashable, T](lambda: value)
        self._nexts = dict[Hashable, T]()

    # [key] => Value
    def __getitem__(self, key: Hashable) -> AggrItem[T]:
        return AggrItem(self, key)

    def get_prev(self, key: Hashable) -> T:
        """Return value at the end of the previous timeslot."""
        return self._prevs[key]

    def get_next(self, key: Hashable) -> T:
        try:
            return self._nexts[key]
        except KeyError:
            return self._prevs[key]

    def set_next(self, key: Hashable, value: T):
        """Schedule update to value in the current timeslot."""
        if value != self.get_next(key):
            self._nexts[key] = value

        # Notify the kernel
        self._set()

    # Variable
    def get_value(self) -> AggrValue[T]:
        """Return present value.

        When performing multiple updates to a variable during the same timeslot,
        this method will always return the value of the *latest* update.
        """
        return AggrValue(self)

    value = property(fget=get_value)

    def changed(self) -> bool:
        return bool(self._nexts)

    def update(self):
        while self._nexts:
            key, value = self._nexts.popitem()
            self._prevs[key] = value


class AggrItem[T](Value[T]):
    """Wrap Aggregate __getitem__."""

    def __init__(self, aggr: Aggregate[T], key: Hashable):
        self._aggr = aggr
        self._key = key

    def get_prev(self) -> T:
        return self._aggr.get_prev(self._key)

    prev = property(fget=get_prev)

    def set_next(self, value: T):
        self._aggr.set_next(self._key, value)

    next = property(fset=set_next)


class AggrValue[T]:
    """Wrap Aggregate value."""

    def __init__(self, aggr: Aggregate[T]):
        self._aggr = aggr

    def __getitem__(self, key: Hashable) -> T:
        return self._aggr.get_next(key)



"""Test basic kernel functionality"""

# pyright: reportAttributeAccessIssue=false
# pyright: reportUnknownMemberType=false
# pyright: reportUnknownVariableType=false

import logging

import pytest
from pytest import LogCaptureFixture

from deltacycle import get_kernel, get_running_kernel, run, set_kernel, sleep, step

logger = logging.getLogger("deltacycle")


async def main(n: int):
    for i in range(n):
        logger.info("%d", i)
        await sleep(1)
    return n


def test_run(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    ret = run(main(42))
    assert ret == 42

    msgs = [(r.time, r.getMessage()) for r in caplog.records]
    assert msgs == [(i, str(i)) for i in range(42)]


def test_irun(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    g = step(main(42))
    try:
        while True:
            next(g)
    except StopIteration as e:
        assert e.value == 42

    msgs = [(r.time, r.getMessage()) for r in caplog.records]
    assert msgs == [(i, str(i)) for i in range(42)]


def test_cannot_run(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    run(main(100))
    kernel = get_kernel()

    # Kernel is already in COMPLETED state
    with pytest.raises(RuntimeError):
        run(kernel=kernel)

    with pytest.raises(RuntimeError):
        list(step(kernel=kernel))


def test_limits(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    run(main(1000), ticks=51)
    kernel = get_running_kernel()
    assert kernel.time() == 50

    run(kernel=kernel, ticks=51)
    assert kernel.time() == 100

    run(kernel=kernel, until=201)
    assert kernel.time() == 200

    # Both ticks & until: first limit to hit
    run(kernel=kernel, ticks=101, until=302)
    assert kernel.time() == 300
    run(kernel=kernel, ticks=102, until=401)
    assert kernel.time() == 400


def test_nocoro():
    with pytest.raises(ValueError):
        run()
    with pytest.raises(ValueError):
        list(step())


def test_get_running_kernel():
    # No kernel
    set_kernel()
    with pytest.raises(RuntimeError):
        get_running_kernel()

    # Kernel is not running
    run(main(42))
    with pytest.raises(RuntimeError):
        get_running_kernel()



"""Queue synchronization primitive."""

from functools import cached_property

from ._kernel_if import KernelIf
from ._task import CreditQ, Task


class Container(KernelIf):
    """Producer / Consumer Resource Container.

    Has both blocking and non-blocking put and get interfaces.
    If capacity is a positive number, the container has *capacity* slots.
    If capacity is zero or a negative number, the container has infinite slots.

    The put interface will block only when it is full.
    The get interface will block only when it is empty.

    An infinite container will never be full.
    Its size is subject only to the machine's memory limitations.
    """

    def __init__(self, capacity: int = 0):
        self._capacity = capacity
        self._cnt: int = 0
        self._getq = CreditQ()
        self._putq = CreditQ()

    def __len__(self) -> int:
        return self._cnt

    @property
    def capacity(self) -> int | None:
        return self._capacity if self._has_capacity else None

    @cached_property
    def _has_capacity(self) -> bool:
        return self._capacity > 0

    def _check_cnt(self):
        assert self._cnt >= 0
        assert not self._has_capacity or self._cnt <= self._capacity

    def _check_n(self, n: int):
        if n < 1:
            raise ValueError(f"Expected n ≥ 1, got {n}")
        if self._has_capacity and n > self._capacity:
            raise ValueError(f"Expected n ≤ {self._capacity}, got {n}")

    def try_put(self, n: int = 1) -> bool:
        self._check_cnt()
        self._check_n(n)

        if self._has_capacity and (self._cnt + n) > self._capacity:
            return False

        # Put credit
        self._cnt += n
        return True

    async def put(self, n: int = 1, priority: int = 0):
        self._check_cnt()
        self._check_n(n)

        if self._has_capacity and (self._cnt + n) > self._capacity:
            task = self._kernel.task()
            self._putq.push((priority, task, n))
            y = await task.switch_coro()
            assert y is None
        else:
            # Put credit
            self._cnt += n

        while self._getq and (self._cnt >= self._getq.peek()):
            # Transfer credit
            task, n = self._getq.pop()
            self._kernel.call_soon(task, args=(Task.Command.RESUME,))
            self._cnt -= n

    def try_get(self, n: int = 1) -> bool:
        self._check_cnt()
        self._check_n(n)

        if self._cnt < n:
            return False

        # Get credit
        self._cnt -= n
        return True

    async def get(self, n: int = 1, priority: int = 0):
        self._check_cnt()
        self._check_n(n)

        if self._cnt < n:
            task = self._kernel.task()
            self._getq.push((priority, task, n))
            y = await task.switch_coro()
            assert y is None
        else:
            # Get credit
            self._cnt -= n

        while self._putq and (self._cnt + self._putq.peek()) <= self._capacity:
            # Transfer credit
            task, n = self._putq.pop()
            self._kernel.call_soon(task, args=(Task.Command.RESUME,))
            self._cnt += n
